/*----------------------------------------------------------------------------------------------------

--

-- File name	 : fifo_pci_interface_controller.V

--					 :

-- Title			 : FIFO PCI Interface Controller

--					 :

-- Purpose		 : Controls data transfer between the PCI interface and the FIFOs

--					 :

-- IncludeFiles :

--					 :

-- Created On	 : August 23, 2001

--					 :

-- Comments		 :

--					 : This is designed based on each burst containing 8 data itesms

--					 : but allowance is made for a write of less than 8 (data will be padded)

--					 :

--					 :

--					 :  For the Spartan II Reference Board

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

--					 :

-- Conventions  : - Port names are upper case.

--					 : - Internal signals and variables are lower case.

--					 : - Active low signals are identified with '_n' or '_N'.

--					 : - constants begin with lower case 'c', rest upper case

--					 : - states begin with lower case 's', rest upper case

--					 : - aliases begin with lower case 'a', rest upper case

--					 :

--					 : Tab setting = 3

--					 : Right Margin = 100

--					 :

-- Limitations  : For use in Xilinx FPGA

--					 :

-- Developers	 : Duncan Allen

----------------------------------------------------------------------------------------------------

-- Revision History :

----------------------------------------------------------------------------------------------------

--	  Version No 	| Author	 	| Changes Made:							| Mod. Date:

--		v1.0			| da			| First write								| 23/8/01

----------------------------------------------------------------------------------------------------*/



// This is a simple synchronous FIFO.

// It assumes that each burst of data will be 8 units long and that

// each burst will be emptied before a new one is loaded









// This module will control the interface between the PCI interface and the FIFOs

// The write interface will be straightforward -- The only complication will be

// if a write of less than 8 units occurs then the remaining space will be padded

// with byte enable off data. If the previous write has not yet emptied the FIFO, the

// design will issue retry.



// On the PCI read side, things are a bit more complicated. On a read, check if there is a read

// pending. If there is, check if the address is correct -- if not issue retry. If the address is correct

// and the FIFO is full then proceed with the read, if the FIFO is not full then retry.



// If there is no read pending then signal to the sdram side to load the FIFO and save the address as well

// as registering the pending read.



`timescale 1 ns/1 ps



module fifo_pci_interface_controller (

		CLK,

		RESET,

		ADDR,

		READ_HIT,

		WRITE_HIT,

		S_DATA,

		SDRAM_READY,

		PCI_READ_UNDERWAY,

		PCI_WRITE_UNDERWAY,

      READ_PRE_LOAD,

		WRITE_COMPLETE,

		READ_COMPLETE,

		BEGIN_WRITE,

		BEGIN_READ,

		READ_ADDRESS,

		WRITE_ADDRESS,

		READ_FLUSH,

		WRITE_FLUSH,

      WRITE_COUNT,  	// The current data count in the write FIFO

      READ_COUNT,    // The current data count in the read FIFO

		READY,

		TERM,

		ABORT

		);



input 			CLK;

input 			RESET;

input [31:0]	ADDR; 				// The current PCI address

input 			READ_HIT;			// Indicates that a PCI hit has occured and it is a read transaction

input 			WRITE_HIT;			// Indicates that a PCI hit has occured and it is a write transaction

input			S_DATA;

input			SDRAM_READY;

output			PCI_READ_UNDERWAY;

output  		PCI_WRITE_UNDERWAY;

output      READ_PRE_LOAD;

input 			WRITE_COMPLETE;

input 			READ_COMPLETE;

output 			BEGIN_WRITE;

output 			BEGIN_READ;

output [24:0]	READ_ADDRESS;

output [24:0]   WRITE_ADDRESS;

output			READ_FLUSH;

output			WRITE_FLUSH;

input [2:0]    WRITE_COUNT;

input [2:0]    READ_COUNT;

output			READY;

output			TERM;

output			ABORT;



reg [24:0]		WRITE_ADDRESS;

reg [24:0]		READ_ADDRESS;

reg				READY;

reg				TERM;

//reg				ABORT;

wire				ABORT;



assign ABORT = 1'b0;



reg				BEGIN_WRITE;

reg				BEGIN_READ;



reg				PCI_READ_UNDERWAY;

reg				PCI_WRITE_UNDERWAY;



reg            READY_pre;              // Preliminary version of S_READY -- It only needs to be delayed for a read but

                                       // it is easier to dealy both



reg				save_write_address;		  	// The write address should be saved

reg				save_read_address;			// The read address should be saved

wire			address_ok;					// The current address matches the pending read address





reg            read_complete_delay_1;     // A delayed version of read complete

reg            read_complete_delay_2;



reg				start_sdram_read;			// Indicates that the SDRAM interface should be told to load the read FIFO

reg				finish_sdram_read;



reg				finish_write;				// Indicates that the PCI portion of the write is complete and the SDRAM part should begin

wire [2:0]      burst_pos;                   // Indicates position inside a standard 8 word burst             

parameter		sSDRAM_INITIALIZE = 1;

parameter 		sIDLE = 2;

parameter		sRETRY = 3;

parameter		sREAD  = 4;

parameter		sWRITE = 5;

parameter      sWRITE_STOP = 6;

parameter      sREAD_STOP = 7;

parameter      sWRITE_SINGLE = 8;

integer 		state;



// For simulation!

/* synthesis translate_off */

reg [(30*8 - 1):0] FSM;

/* synthesis translate_on */



// Save the write address

always@(posedge CLK or posedge RESET)

	begin

		if (RESET)

			WRITE_ADDRESS <= 25'h0000000;

		else

			if (save_write_address)

				WRITE_ADDRESS <= ADDR [24:0];

	end



// Burst pos is the address

assign burst_pos = ADDR [4:2]; 

	

// Save the read address

always@(posedge CLK or posedge RESET)

	begin

		if (RESET) 

			READ_ADDRESS <= 25'h0000000;

		else

			if (save_read_address)

				READ_ADDRESS <= ADDR [24:0];

	end



assign address_ok = (READ_ADDRESS == ADDR [24:0]);



// The main state machine

always @ (posedge CLK or posedge RESET)

	begin

		if (RESET)

			begin

			READY_pre <= 1'b0;

			TERM <= 1'b0;

//			ABORT <= 1'b0;

			PCI_READ_UNDERWAY <= 1'b0;

			PCI_WRITE_UNDERWAY <= 1'b0;



			save_read_address <= 1'b0;

			save_write_address <= 1'b0;

			finish_write <= 1'b0;

			state <= sSDRAM_INITIALIZE;

			start_sdram_read <= 1'b0;

			finish_sdram_read <= 1'b0;

			end

		else

			case (state)

				sSDRAM_INITIALIZE : begin

					if (READ_HIT || WRITE_HIT) 	// If there is an attempt to read or write before the SDRAM is ready, signal a RETRY

						begin

						state <= sRETRY;

						TERM <= 1'b1;

						end

					else if (SDRAM_READY)

						begin

						state <= sIDLE;

						TERM <= 1'b0;

						end

					else

						begin

						state <= sSDRAM_INITIALIZE;

						TERM <= 1'b0;

						end

				end



				sIDLE : begin

					finish_write <= 1'b0;

					finish_sdram_read <= 1'b0;

					if (WRITE_HIT)

						if (BEGIN_WRITE)	// If the previous write hasn't finished

							begin

							state <= sRETRY;

							TERM <= 1'b1;

							end

						else if (burst_pos == 3'b000)

							begin

							state <= sWRITE;

							READY_pre <= 1'b1;

							save_write_address <= 1'b1;

							end	  

						else 

							begin

							state <= sWRITE_SINGLE;

							READY_pre <= 1'b1;

							save_write_address <= 1'b1;

							end

					else if (READ_HIT)

							begin

							if (~BEGIN_READ)		// There is no other read to be serviced

								begin

								state <= sRETRY;

								TERM <= 1'b1;

								save_read_address <= 1'b1;

								start_sdram_read <= 1'b1;

								end

							else

								if (address_ok && read_complete_delay_2) // If the addresses match and the FIFO is full

								begin

                        PCI_READ_UNDERWAY <= 1'b1;

								state <= sREAD;

								READY_pre <= 1'b1;       // Hold off for one clock to allow for FIFO latency

								end

							else

								begin

								TERM <= 1'b1;

								state <= sRETRY;

								end

							end

					else

                  begin

                  PCI_READ_UNDERWAY <= 1'b0;

				  PCI_WRITE_UNDERWAY <= 1'b0;

						state <= sIDLE;

						TERM <= 1'b0;

						READY_pre <= 1'b0;

//						ABORT <= 1'b0;

                  end

				end



				sRETRY : begin

					save_read_address <= 1'b0;

					save_write_address <= 1'b0;

					start_sdram_read <= 1'b0;

					if (S_DATA)			// The transaction still has not finished

						begin

							TERM <= 1'b1;

							state <= sRETRY;



						end

					else

						begin

							TERM <= 1'b0;

							if (SDRAM_READY)

								state <= sIDLE;

							else

								state <= sSDRAM_INITIALIZE;



						end

				end



				sWRITE : begin				// The PCI_WRITE is underway

					PCI_WRITE_UNDERWAY <= 1'b1;

					save_write_address <= 1'b0;

					if (S_DATA)

                  begin

                  if (WRITE_COUNT == 3'b100)    // There will be more than 8 data items if the write is not stopped

                     begin

					 state <= sWRITE_STOP;

                     TERM <= 1'b1;

                     READY_pre <= 1'b1;

                     finish_write <= 1'b0;

                     end



                  else                             // Continuue on with the write

							begin

							state <= sWRITE;

							READY_pre <= 1'b1;

							finish_write <= 1'b0;

							end

                  end

					else      // Initiator has stopped the transaction

						begin

						state <= sIDLE;

						READY_pre <= 1'b0;

						finish_write <= 1'b1;

						end

				end

			

			sWRITE_SINGLE : begin           // A single transaction write becuase it was not on an 8 dword boundary

				PCI_WRITE_UNDERWAY <= 1'b1;

				save_write_address <= 1'b0;

				if (S_DATA)				 

					begin

					state <= sWRITE_SINGLE;

					TERM <= 1'b1;

					READY_pre <= 1'b1;

					finish_write <= 1'b0;

					end

				else

					begin

					state <= sIDLE;

					READY_pre <= 1'b0;

					finish_write <= 1'b1;

					end

			end

			sREAD : begin

				PCI_READ_UNDERWAY <= 1'b1;

				save_read_address <= 1'b0;

				if (S_DATA)

               begin

               if (READ_COUNT == 3'b000)  // There will be more  8 items read if the read is not stopped

               	begin

                  state <= sREAD_STOP;

                  READY_pre <= 1'b1;

                  TERM <= 1'b1;

                  end

               else

						begin

						state <= sREAD;

						READY_pre <= 1'b1;

         			end

               end

				else

					begin

					state <= sIDLE;

					READY_pre <= 1'b0;

					finish_sdram_read <= 1'b1;

					end

			end



         sWRITE_STOP: begin

            PCI_WRITE_UNDERWAY <= 1'b1;

            if (S_DATA)     // The transaction has not yet finished

               begin

               state <= sWRITE_STOP;

               READY_pre <= 1'b1;

               TERM <= 1'b1;

               finish_write <= 1'b0;

               end

            else

               begin

               state <= sIDLE;

               READY_pre <= 1'b0;

               TERM <= 1'b0;

               finish_write <= 1'b1;

               end

         end



         sREAD_STOP : begin

            PCI_READ_UNDERWAY <= 1'b1;

            if (S_DATA)    // The transaction has not yet finished

               begin

               state <= sREAD_STOP;

               READY_pre <= 1'b1;

               TERM <= 1'b1;

               finish_sdram_read <= 1'b0;

               end

            else

               begin

               state <= sIDLE;

               READY_pre <= 1'b0;

               TERM <= 1'b0;

               finish_sdram_read <= 1'b1;

               end

         end



			default : begin

				state <= sIDLE;

			end

			endcase

		end





// Tell the SDRAM interface to begin writing

// Clear the BEGIN_WRITE signal when the SDRAM controller indicates that it is finished

always @(posedge CLK or posedge RESET)

	begin

		if (RESET)

		 	BEGIN_WRITE <= 1'b0;

		else

			begin

			if (finish_write)

				BEGIN_WRITE <= 1'b1;

			else if (WRITE_COMPLETE)

				BEGIN_WRITE <= 1'b0;

			end

end



always @(posedge CLK or posedge RESET)

   begin

   if (RESET)

      begin

      read_complete_delay_1 <= 1'b0;

      read_complete_delay_2 <= 1'b0;

      end

   else

      begin

      read_complete_delay_1 <= READ_COMPLETE;

      read_complete_delay_2 <= read_complete_delay_1;

      end

   end



assign READ_PRE_LOAD = READ_COMPLETE && !read_complete_delay_2;



assign WRITE_FLUSH = WRITE_COMPLETE;



assign READ_FLUSH = finish_sdram_read;



// Control the read FIFO

always @ (posedge CLK or posedge RESET)

	begin

		if (RESET)

			BEGIN_READ <= 1'b0;

		else

			begin

			if(start_sdram_read)

				BEGIN_READ <= 1'b1;

			else if (finish_sdram_read)

				BEGIN_READ <= 1'b0;



			end

	end



// Delay the S_READY signal

always @ (posedge CLK or posedge RESET)

   begin

   if (RESET)

      READY <= 1'b0;

   else

      READY <= READY_pre;

   end



//assign READY = READY_pre;



// For simulation!

/* synthesis translate_off */

always @(state)

begin

   case(state)

			1:     FSM = "sSDRAM_INITIALIZE";

      	2:     FSM = "sIDLE";

      	3:     FSM = "sRETRY";

      	4:     FSM = "sREAD";

      	5:     FSM = "sWRITE";

         6:     FSM = "sWRITE_STOP";

         7:     FSM = "sREAD_STOP";

		 	8:     FSM = "sWRITE_SINGLE";

         default:      FSM = "UNKNOWN";

   endcase

end

/* synthesis translate_on */



endmodule

