----------------------------------------------------------------------------------------------------

--             :

-- File name   : PCI_HOST.VHD

--             :

-- Title       : PCI host VHDL model

--             :

-- IncludeFiles:

--             :

-- Created On  :

--             :

-- Comments    :

--             :

--             :

--             :

--             :

--             :

--             :

--             :

--             :

-- Conventions : - Port names are 'UPPER' case.

--             : - Internal signals and variables are 'lower' case.

--             : - Active low signals are identified with '_n' or '_N'.

--             : - states begin with lower case 's', rest upper case

--             : - aliases begin with lower case 'a', rest upper case

--             : - instances begin with lower case 'i', rest upper case

--             :

--             : Tab setting = 3

--             : Right Margin = 100

--             :

-- Limitations : For use in Xilinx FPGA

--             :

-- Developers  : Majid Kalsi

----------------------------------------------------------------------------------------------------

-- Revision History :

----------------------------------------------------------------------------------------------------

--   Version No | Author    | Changes Made:                                            | Mod. Date:

--     v1.0     | mk        | First write                                              | 03/27/00

----------------------------------------------------------------------------------------------------





library IEEE;

use IEEE.STD_LOGIC_1164.all;

use IEEE.STD_LOGIC_UNSIGNED.all;

use IEEE.STD_LOGIC_ARITH.all;

use IEEE.STD_LOGIC_TEXTIO.all;

use STD.TEXTIO.all;

use WORK.UTILS_PKG.all;





entity PCI_Host is

generic( LogFile : STRING  := "HOST.LOG");

port(

   -- command/control interface

   INITIATOR_CMD  : in integer;

   TARGET_CMD     : in integer;

   END_SIM        : in boolean;

   INSERT_WAITS   : in boolean;

   CMD_ACK        : out boolean;

   BUSY           : out boolean;

   -- PCI signals

   CLK            : in     std_logic;

   RESET_N        : in     std_logic;

   INTR_A         : in     std_logic;

   REQ_N          : out    std_logic;

   GNT_N          : in     std_logic;

   FRAME_N        : inout  std_logic;

   IRDY_N         : inout  std_logic;

   IDSEL          : out    std_logic;

   DEVSEL_N       : inout  std_logic;

   TRDY_N         : inout  std_logic;

   STOP_N         : inout  std_logic;

   CBE            : inout  std_logic_vector(3 downto 0);

   AD             : inout  std_logic_vector(31 downto 0);

   PAR            : inout  std_logic

);

end PCI_Host;





----------------------------------------------------------------------------------------------------

architecture PCI_Host_arch of PCI_Host is

----------------------------------------------------------------------------------------------------





----------------------------------------------------------------------------------------------------

-- define TYPES

----------------------------------------------------------------------------------------------------

type RWTYPE is (READ, WRITE);

type SPACE_TYPE is (CFG, IO, MEM);

type STATUS_TYPE is (NORMAL, DISCONNCT, RETRY, CHECK, DEAD, T_ABORT, M_ABORT, NOTHING);

type CYCLE_TYPE is (IDLE, WRITE_CFG, READ_CFG, WRITE_IO, READ_IO, WRITE_MEM, READ_MEM);

type STATE_TYPE is (IDLE, RESET, INIT, BUSY_LED_ON, BUSY_LED_OFF,

                    FULL_XFER, PARTIAL_XFER, PERR_XFER,

						  T_ABORT_XFER, M_ABORT_XFER, RETRY_XFER, 

						  FINISH, ISR);



----------------------------------------------------------------------------------------------------

-- define constants

----------------------------------------------------------------------------------------------------



-- Host's Initiator commands

constant NONE                 : integer := 0;

constant LED_ON               : integer := 1;

constant LED_OFF              : integer := 2;



-- Host's Target commands

constant TARGET_ABORT         : integer := 1;

constant MASTER_ABORT         : integer := 2;

constant PARITY_ERROR         : integer := 3;

constant PARTIAL_BURST        : integer := 4;

constant RETRY_BURST          : integer := 5;



-- timing parameters

constant Tdel         : time := 6.0 ns;



-- PCI Configuration Addresses

constant CFG_CSR     :  std_logic_vector(31 downto 0) := X"00000004";

constant BAR0        :  std_logic_vector(31 downto 0) := X"00000010";

constant BAR1        :  std_logic_vector(31 downto 0) := X"00000014";

constant BAR2        :  std_logic_vector(31 downto 0) := X"00000018";

constant LAT_TIMER   :  std_logic_vector(31 downto 0) := X"0000000C";

constant HOST_BASE   :  std_logic_vector(31 downto 0) := X"C0000000";





-- PCI Command codes

constant INT_ACK     : std_logic_vector(3 downto 0) := "0000";

constant SPECIAL     : std_logic_vector(3 downto 0) := "0001";

constant IO_READ     : std_logic_vector(3 downto 0) := "0010";

constant IO_WRITE    : std_logic_vector(3 downto 0) := "0011";

constant MEM_READ    : std_logic_vector(3 downto 0) := "0110";

constant MEM_WRITE   : std_logic_vector(3 downto 0) := "0111";

constant CFG_READ    : std_logic_vector(3 downto 0) := "1010";

constant CFG_WRITE   : std_logic_vector(3 downto 0) := "1011";

constant READ_MULT   : std_logic_vector(3 downto 0) := "1100";

constant DUAL_ADDR   : std_logic_vector(3 downto 0) := "1101";

constant READ_LINE   : std_logic_vector(3 downto 0) := "1110";

constant WRITE_INV   : std_logic_vector(3 downto 0) := "1111";    -- memory write invalidate



constant BAR0_BASE   : std_logic_vector(31 downto 0) := X"B0000000";



-- Byte Enable Masks

constant EN_ALL         : std_logic_vector(3 downto 0) := "0000";

constant EN_BYTE0       : std_logic_vector(3 downto 0) := "1110";

constant EN_BYTE1       : std_logic_vector(3 downto 0) := "1101";

constant EN_BYTE2       : std_logic_vector(3 downto 0) := "1011";

constant EN_BYTE3       : std_logic_vector(3 downto 0) := "0111";

constant EN_MSWORD      : std_logic_vector(3 downto 0) := "0011";

constant EN_LSWORD      : std_logic_vector(3 downto 0) := "1100";



----------------------------------------------------------------------------------------------------

-- define signals

----------------------------------------------------------------------------------------------------

signal reg_idsel      : std_logic;

signal reg_req_n      : std_logic;

signal reg_frame_n    : std_logic;

signal reg_irdy_n     : std_logic;



signal reg_trdy_n     : std_logic;

signal reg_stop_n     : std_logic;

signal reg_devsel_n   : std_logic;



-- Host Initiator signals

signal init_oe     : std_logic;

signal init_par_oe : std_logic;

signal init_par    : std_logic;

signal init_ad_oe  : std_logic;

signal init_ad     : std_logic_vector(31 downto 0);

signal init_cbe_oe : std_logic;

signal init_cbe    : std_logic_vector(3 downto 0);

signal init_cycle  : CYCLE_TYPE;

signal init_state  : STATE_TYPE;

signal init_busy   : boolean := false;

signal init_ack    : boolean := false;



-- Host Target signals

signal targ_oe     : std_logic;

signal targ_par_oe : std_logic;

signal targ_par    : std_logic;

signal targ_ad_oe  : std_logic;

signal targ_ad     : std_logic_vector(31 downto 0);

signal targ_cbe_oe : std_logic;

signal targ_cbe    : std_logic_vector(3 downto 0);

signal targ_state  : STATE_TYPE;

signal targ_ack    : boolean := false;

signal targ_busy   : boolean := false;

signal gen_perr    : std_logic;



-- random number

signal rnd_num        : real;



--file Log_file : text is out LogFile;

file log_file:    TEXT open write_mode is LogFile;



----------------------------------------------------------------------------------------------------

-- define Global procedures

----------------------------------------------------------------------------------------------------



-- procedure to print messages to stdout

procedure print(s: in string) is

variable l1: line;

variable l2: line;

variable str : string(1 downto 1) := " ";

begin

   Write(l1,now);

   Write(l1, str);

   Write(l1, s);

   WriteLine(output,l1);



   Write(l2,now);

   Write(l2, str);

   Write(l2, s);

   WriteLine(Log_file,l2);

end print;



-- Procedure for waiting

procedure Intercycle_Gap is

begin

-- wait until rising_edge(CLK);

-- wait until rising_edge(CLK);

-- wait until rising_edge(CLK);

-- wait until rising_edge(CLK);

end Intercycle_gap;





----------------------------------------------------------------------------------------------------

begin    -- begin Architecture

----------------------------------------------------------------------------------------------------

-- Define port hookup.

FRAME_N  <= reg_frame_n  after Tdel when init_oe = '1' else 'Z' after Tdel;

IRDY_N   <= reg_irdy_n   after Tdel when init_oe = '1' else 'Z' after Tdel;

TRDY_N   <= reg_trdy_n   after Tdel when targ_oe = '1' else 'Z' after Tdel;

STOP_N   <= reg_stop_n   after Tdel when targ_oe = '1' else 'Z' after Tdel;

DEVSEL_N <= reg_devsel_n after Tdel when targ_oe = '1' else 'Z' after Tdel;



AD  <= init_ad  after Tdel when init_ad_oe = '1'  else (others => 'Z') after Tdel;

CBE <= init_cbe after Tdel when init_cbe_oe = '1' else (others => 'Z') after Tdel;

PAR <= init_par after Tdel when init_par_oe = '1' else 'Z' after Tdel;



AD  <= targ_ad  after Tdel when targ_ad_oe = '1'  else (others => 'Z') after Tdel;

CBE <= targ_cbe after Tdel when targ_cbe_oe = '1' else (others => 'Z') after Tdel;

PAR <= targ_par after Tdel when targ_par_oe = '1' else 'Z' after Tdel;



IDSEL    <= 'L' when RESET_N = '0' else reg_idsel after Tdel;

REQ_N    <= 'H' when RESET_N = '0' else reg_req_n after Tdel;



CMD_ACK <= true when init_ack  or targ_ack  else false;

BUSY    <= true when init_busy or targ_busy else false;







-- PCI Parity Generation

process(CLK, RESET_N)

variable count : integer;

variable target_par : std_logic;

begin

   if RESET_N = '0' then

      init_par <= '0';

      targ_par <= '0';

      count := 0;

   elsif rising_edge(CLK) then

      count := count + 1;

      init_par <= init_cbe(0) xor init_cbe(1) xor init_cbe(2) xor init_cbe(3) xor

                  init_ad( 0) xor init_ad( 1) xor init_ad( 2) xor init_ad( 3) xor

                  init_ad( 4) xor init_ad( 5) xor init_ad( 6) xor init_ad( 7) xor

                  init_ad( 8) xor init_ad( 9) xor init_ad(10) xor init_ad(11) xor

                  init_ad(12) xor init_ad(13) xor init_ad(14) xor init_ad(15) xor

                  init_ad(16) xor init_ad(17) xor init_ad(18) xor init_ad(19) xor

                  init_ad(20) xor init_ad(21) xor init_ad(22) xor init_ad(23) xor

                  init_ad(24) xor init_ad(25) xor init_ad(26) xor init_ad(27) xor

                  init_ad(28) xor init_ad(29) xor init_ad(30) xor init_ad(31);



      target_par := targ_cbe(0) xor targ_cbe(1) xor targ_cbe(2) xor targ_cbe(3) xor

                    targ_ad( 0) xor targ_ad( 1) xor targ_ad( 2) xor targ_ad( 3) xor

                    targ_ad( 4) xor targ_ad( 5) xor targ_ad( 6) xor targ_ad( 7) xor

                    targ_ad( 8) xor targ_ad( 9) xor targ_ad(10) xor targ_ad(11) xor

                    targ_ad(12) xor targ_ad(13) xor targ_ad(14) xor targ_ad(15) xor

                    targ_ad(16) xor targ_ad(17) xor targ_ad(18) xor targ_ad(19) xor

                    targ_ad(20) xor targ_ad(21) xor targ_ad(22) xor targ_ad(23) xor

                    targ_ad(24) xor targ_ad(25) xor targ_ad(26) xor targ_ad(27) xor

                    targ_ad(28) xor targ_ad(29) xor targ_ad(30) xor targ_ad(31);



		-- if generating parity error

		if gen_perr = '1' then

         if (count mod 3) = 2 then

         	targ_par <= not target_par;

         else

         	targ_par <= target_par;

         end if;

      else

         targ_par <= target_par;

      end if;

   end if;

end process;







----------------------------------------------------------------------------------------------------

-- PCI Host Initiator

----------------------------------------------------------------------------------------------------



HOST_INITIATOR : process



----------------------------------------------------------------------------------------------------

-- Procedure for monitoring the data transfer.

----------------------------------------------------------------------------------------------------

procedure Xfer_Status (write_read  : in RWTYPE;

                       return_stat : out STATUS_TYPE) is

variable devsel_cnt : integer;

variable trdy_cnt   : integer;



begin

   devsel_cnt := 0;

   trdy_cnt   := 0;



 -- if device has not been selected

   while ((To_X01(DEVSEL_N) = '1') and (devsel_cnt < 10)) loop

      wait until rising_edge(CLK);

      devsel_cnt := devsel_cnt + 1;

   end loop;



   while ((To_X01(TRDY_N)='1') and (To_X01(STOP_N)='1') and (trdy_cnt < 16) and (devsel_cnt < 10)) loop

       -- if TRDY_N deasserted, no abort or error so continue

      trdy_cnt := trdy_cnt + 1;

      wait until rising_edge(CLK);

   end loop;



   if (devsel_cnt < 10) then

      if (trdy_cnt <= 16) then

         if ( To_X01(TRDY_N)='0' and To_X01(STOP_N)='1' ) then

            if (write_read = WRITE) then

               print("    HOST-->SLAVE: Normal Termination, Data Transferred" & LF);

            else

               print("    HOST<--SLAVE: Normal Termination, Data Transferred");

            end if;

            return_stat := NORMAL;

         elsif ( To_X01(TRDY_N)='0' and To_X01(STOP_N)='0' ) then

            if (write_read = WRITE) then

               print("    HOST-->SLAVE: Disconnect, Data Transferred" & LF);

            else

               print("    HOST<--SLAVE: Disconnect, Data Transferred");

            end if;

            return_stat := DISCONNCT;

         elsif ( To_X01(TRDY_N)='1' and To_X01(STOP_N)='0' and To_X01(DEVSEL_N)='0' ) then

            if (write_read = WRITE) then

               print("    HOST-->SLAVE: Retry, No Data Transferred" & LF);

            else

               print("    HOST<--SLAVE: Retry, No Data Transferred");

            end if;

            return_stat := RETRY;

         elsif ( To_X01(TRDY_N)='1' and To_X01(STOP_N)='0' and To_X01(DEVSEL_N)='1' ) then

            if (write_read = WRITE) then

               print("    HOST-->SLAVE: Target Abort, No Data Transferred" & LF);

            else

               print("    HOST<--SLAVE: Target Abort, No Data Transferred");

            end if;

            return_stat := T_ABORT;

         elsif ( To_X01(TRDY_N)='1' and To_X01(STOP_N)='1' ) then

            assert false

            report "    Check Transfer Procedure" & LF

            severity error;

            return_stat := CHECK;

         end if;

      else

         assert false

         report "    No Target Response" & LF

         severity warning;

         return_stat := DEAD;

      end if;

   else

      print("    Master Abort, No Data Transferred");

--      assert false

--      report "    Master Abort" & LF

--      severity warning;

      return_stat := M_ABORT;

   end if;



end Xfer_status;





----------------------------------------------------------------------------------------------------

-- Procedure for reading from target

----------------------------------------------------------------------------------------------------

procedure read_target (space   : in SPACE_TYPE;

                       address : in std_logic_vector(31 downto 0);

                       data    : out std_logic_vector(31 downto 0)) is

variable status_code : STATUS_TYPE;

variable cmd : std_logic_vector(3 downto 0);



begin

   case space is

      when CFG =>

         print("    Reading Configuration register...");

         cmd := CFG_READ;

         init_cycle <= READ_CFG;

      when IO  =>

         print("    Reading IO register...");

         cmd := IO_READ;

         init_cycle <= READ_IO;

      when MEM =>

         print("    Reading Memory...");

         cmd := MEM_READ;

         init_cycle <= READ_MEM;

   end case;

   print("    CBE = " & vec2bstr(cmd,4));

   print("    Address = " & vec2hstr(address,8));

--   wait until rising_edge(CLK);



   -- assert request and wait for grant

   reg_req_n <= '0';

   while(GNT_N = '1') loop

      wait until rising_edge(CLK);

   end loop;

   reg_req_n <= '1';



   -- address phase

   reg_frame_n <= '0';        -- assert frame

   reg_irdy_n <= '1';

   init_oe <= '1';

   init_ad <= address;        -- assert address

   init_ad_oe <= '1';

   init_cbe <= cmd;           -- assert command

   init_cbe_oe <= '1';



   -- assert IDSEL only during configuration accesses

   if space = CFG then

      reg_idsel <= '1';

   else

      reg_idsel <= '0';

   end if;

   wait until rising_edge(CLK);



   -- IRDY_N phase

   reg_idsel <= '0';          -- de-assert idsel, if asserted

   reg_frame_n <= '1';        -- de-assert frame

   reg_irdy_n <= '0';         -- assert irdy

   init_cbe <= "0000";        -- assert byte enables

   init_ad_oe <= '0';         -- tri-state AD drivers

   init_par_oe <= '1';        -- drive par

   wait until rising_edge(CLK);

   init_par_oe <= '0';



   -- wait for data

   Xfer_Status(READ, status_code);

   if status_code = NORMAL or status_code = DISCONNCT then

      data := AD;

      print("    Data Read = " & vec2hstr(AD,8) & LF);

   else

      data := (others => 'Z');

   end if;



   -- terminate access

   reg_irdy_n <= '1';         -- de-assert irdy

   init_cbe_oe <= '0';        -- tri-state CBE drivers

   wait until rising_edge(CLK);



   init_oe <= '0';            -- tri-state initiator signals

   wait until rising_edge(CLK);



   init_cycle <= IDLE;

   Intercycle_Gap;

   if status_code = RETRY then

      print("    RETRYING TARGET READ...");

      Read_Target(space, address, data);

   end if;

end read_target;





----------------------------------------------------------------------------------------------------

-- Procedure for writing to target

----------------------------------------------------------------------------------------------------

procedure write_target (space : in SPACE_TYPE;

                        be : in std_logic_vector(3 downto 0);

                        address : in std_logic_vector(31 downto 0);

                        data : in std_logic_vector(31 downto 0)) is

variable status_code : STATUS_TYPE;

variable cmd : std_logic_vector(3 downto 0);

begin

   case space is

      when CFG =>

         print("    Writing Configuration register...");

         cmd := CFG_WRITE;

         init_cycle <= WRITE_CFG;

      when IO  =>

         print("    Writing IO register...");

         cmd := IO_WRITE;

         init_cycle <= WRITE_IO;

      when MEM =>

         print("    Writing Memory...");

         cmd := MEM_WRITE;

         init_cycle <= WRITE_MEM;

   end case;

   print("    CBE = " & vec2bstr(cmd,4));

   print("    Address = " & vec2hstr(address,8));

   print("    Data = " & vec2hstr(data,8));

--   wait until rising_edge(CLK);



   -- assert request and wait for grant

   reg_req_n <= '0';

   while(GNT_N = '1') loop

      wait until rising_edge(CLK);

   end loop;

   reg_req_n <= '1';



   -- address phase

   reg_frame_n <= '0';        -- assert frame

   reg_irdy_n <= '1';

   init_oe <= '1';

   init_ad <= address;        -- assert address

   init_ad_oe <= '1';

   init_cbe <= cmd;           -- assert command

   init_cbe_oe <= '1';



   -- assert IDSEL only during configuration accesses

   if space = CFG then

      reg_idsel <= '1';

   else

      reg_idsel <= '0';

   end if;

   wait until rising_edge(CLK);



   -- IRDY_N phase

   reg_idsel <= '0';          -- de-assert idsel, if asserted

   reg_frame_n <= '1';        -- de-assert frame

   reg_irdy_n <= '0';         -- assert irdy

   init_cbe <= be;            -- assert byte enables

   init_ad <= data;           -- assert data

   init_par_oe <= '1';        -- drive par

   wait until rising_edge(CLK);





   -- wait for data and terminate access

   Xfer_Status(WRITE, status_code);

   wait for 1 ns;

   reg_irdy_n <= '1';         -- de-assert irdy

   init_ad_oe <= '0';         -- tri-state AD drivers

   init_cbe_oe <= '0';        -- tri-state CBE drivers

   wait until rising_edge(CLK);



   init_oe <= '0';            -- tri-state initiator drivers

   init_par_oe <= '0';        -- tri-state PAR driver

   wait until rising_edge(CLK);



   init_cycle <= IDLE;

   Intercycle_Gap;

   if status_code = RETRY then

      print("    Retrying Target Write...");

      Write_Target(space, be, address, data);

   end if;

end write_target;



----------------------------------------------------------------------------------------------------

-- Procedure for Initializing Everything

----------------------------------------------------------------------------------------------------

procedure Initialize is

variable rdata : std_logic_vector(31 downto 0);

begin

   init_busy <= true;

   init_state <= RESET;

   init_oe <= '0';

   reg_req_n <= '1';

   reg_frame_n <= '1';

   reg_irdy_n <= '1';

   reg_idsel <= '0';



   init_ad <= (others => '0');

   init_cbe <= (others => '0');

   init_ad_oe <= '0';

   init_cbe_oe <= '0';

   init_par_oe <= '0';



   wait until RESET_N = '1';



   wait until rising_edge(CLK);

   wait until rising_edge(CLK);

   wait until rising_edge(CLK);



   print("[HOST INITIATOR] Begin Initialization...");

   init_state <= INIT;



   -- Setup command register to enable:

   -- I/O and memory access

   -- Parity error response

   -- System error enable

   -- Clear status register

   print("[HOST INITIATOR] Setup PCI Configuration Register...");

   write_target(CFG, EN_ALL, CFG_CSR, X"FFFF01C2");

   read_target(CFG, CFG_CSR, rdata);

   assert(rdata = X"020001C2")

   report "PCI Configuration Command register error"

   severity error;



   -- Verify BAR0 is MEMORY and size of 1M

   print("[HOST INITIATOR] Verifying BAR0 is 1M...");

   write_target(CFG, EN_ALL, BAR0, X"FFFFFFFF");

   read_target(CFG, BAR0, rdata);

   -- check result

   assert rdata = X"FFF00008"

   report " PCI Configuration base IO address range error!" & LF &

          " Expected: " & vec2hstr(X"FFF00008", 8) & "  Received: " & vec2hstr(rdata, 8)

   severity error;



   -- Define starting Address for BARO address space

   print("[HOST INITIATOR] defining start of BAR0 address space...");

   write_target(CFG, EN_ALL, BAR0, BAR0_BASE);



   -- Setup Latency Timer

   print("[HOST INITIATOR] Setting Latency Timer to Maximum...");

   write_target(CFG, EN_BYTE1, LAT_TIMER, X"0000FF00");



   -- Verify Latency Timer is Setup

   print("[HOST INITIATOR] Verifying Latency Timer Setup...");

   read_target(CFG, LAT_TIMER, rdata);

   -- check result

   rdata := rdata and X"0000FF00";

   assert rdata = X"0000FF00"

   report " PCI Configuration Latency Timer Setup error!" & LF &

          " Expected: " & vec2hstr(X"0000FF00", 8) & "  Received: " & vec2hstr(rdata, 8)

   severity error;



   init_busy <= false;

   init_state <= IDLE;

   print("[HOST INITIATOR] Initialization End" & LF);

end Initialize;





----------------------------------------------------------------------------------------------------

-- Procedure for turning off the led

----------------------------------------------------------------------------------------------------

procedure turn_led_off is

variable data : std_logic_vector(31 downto 0);

begin

   init_busy <= true;

   init_state <= BUSY_LED_OFF;

   print("[HOST INITIATOR] Turning OFF the LED...");



   --write a single dword to turn off the led

   print(" ");

   print("    WRITE a single dword to bar0");

   write_target(MEM, EN_ALL, BAR0_BASE, X"FFFFFFF0");



   print("[HOST INITIATOR] Finished Turn LED Off" & LF);

   init_busy <= false;

   init_state <= IDLE;



end turn_led_off;



----------------------------------------------------------------------------------------------------

-- Procedure for turning on the led

----------------------------------------------------------------------------------------------------

procedure turn_led_on is

variable data : std_logic_vector(31 downto 0);

begin

   init_busy <= true;

   init_state <= BUSY_LED_ON;

   print("[HOST INITIATOR] Turning OFF the LED...");



   --write a single dword to turn on the led

   print(" ");

   print("    WRITE a single dword to bar0");

   write_target(MEM, EN_ALL, BAR0_BASE, X"FFFFFFF1");



   print("[HOST INITIATOR] Finished Turn LED On" & LF);

   init_busy <= false;

   init_state <= IDLE;



end turn_led_on;



----------------------------------------------------------------------------------------------------

begin  -- Main Initiator Loop

----------------------------------------------------------------------------------------------------

   wait on CLK, RESET_N;



   if END_SIM then

      init_state <= FINISH;

      file_close(log_file);

      wait;

   else

      if RESET_N = '0' then

         init_state <= RESET;

         Initialize;

      else

         -- if HOST target is busy then wait

         if targ_busy then

            wait until not targ_busy;

         end if;



         case INITIATOR_CMD is

            when NONE =>



            when LED_ON =>

               init_ack <= true;

               wait for 30 ns;

               init_ack <= false;

               turn_led_on;



				when LED_OFF =>

               init_ack <= true;

               wait for 30 ns;

               init_ack <= false;

               turn_led_off;               



				when others =>

               null;

         end case;

      end if;

   end if;

end process HOST_INITIATOR;





----------------------------------------------------------------------------------------------------

-- PCI Host Target

----------------------------------------------------------------------------------------------------

HOST_TARGET : process



variable AD_in  : std_logic_vector(31 downto 0);

variable cbe_in : std_logic_vector(3 downto 0);

variable read   : boolean;



----------------------------------------------------------------------------------------------------

-- Procedure for Master Abort

----------------------------------------------------------------------------------------------------

procedure mabort_transfer is

begin



   targ_busy <= true;

   targ_state <= M_ABORT_XFER;



   print("[HOST TARGET] Performing Master Abort...");



   -- wait for IRDY assert

	if To_X01(IRDY_N) = '1' then

      wait until To_X01(IRDY_N) = '0';

   end if;



   -- wait for IRDY de-assert

	if To_X01(IRDY_N) = '0' then

      wait until To_X01(IRDY_N) = '1';

   end if;



	print("[HOST TARGET] End of Master Abort" & LF);

   targ_busy <= false;

   targ_state <= IDLE;

end mabort_transfer;





----------------------------------------------------------------------------------------------------

-- Procedure for Target Abort

-- write 5 data phase(s) and then target abort

----------------------------------------------------------------------------------------------------

procedure tabort_transfer is

variable index : integer;

variable count : integer := 0;

begin



   targ_busy <= true;

   targ_state <= T_ABORT_XFER;

   index := vec2int(AD_in(27 downto 2));



   print("[HOST TARGET] Performing Target Abort...");

   print("    Address: " & vec2hstr(AD_in,8));



   -- enable driving of target signals

   targ_oe <= '1';



   -- wait for IRDY_N

   while To_X01(IRDY_N) = '1' loop

   	wait until rising_edge(CLK);

   end loop;



   -- assert DEVSEL_N

   reg_devsel_n <= '0';

   wait until rising_edge(CLK);



	-- assert TRYDY_N

   reg_trdy_n <= '0';

   targ_cbe <= CBE;



	-- if read cycle

   if read then

   	targ_ad_oe <= '1';

      wait for Tdel;

   	print("    Reading from host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   	targ_par_oe <= '1';

   else

      wait for Tdel;

   	print("    Writing to host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   end if;



	index := index + 1;

	-- loop 5 times or until frame asserted

	while count < 4 and To_X01(FRAME_N) = '0' and To_X01(IRDY_N) = '0' loop

      targ_cbe <= CBE;

      if read then

         wait for Tdel;

      else

         wait for Tdel;

      end if;

      wait until rising_edge(CLK);

      print("    Data Transferred: " & vec2hstr(AD,8));

      count := count + 1;

  		index := index + 1;

   end loop;



   -- if frame not asserted after 5 data phases

   if To_X01(FRAME_N) = '0' then

      -- assert target abort signals

      reg_devsel_n <= '1';

      reg_trdy_n <= '1';

      reg_stop_n <= '0';

      targ_cbe <= CBE;

      targ_ad_oe <= '0';

   	wait until rising_edge(CLK);

      wait until rising_edge(CLK);

   end if;



   -- de-assert signals

   reg_devsel_n <= '1';

  	reg_trdy_n <= '1';

   reg_stop_n <= '1';

  	targ_ad_oe <= '0';

   targ_par_oe <= '0';

   wait until rising_edge(CLK);



   -- disable driving of target signals

   targ_oe <= '0';



   print("[HOST TARGET] End of Target Abort" & LF);

   targ_busy <= false;

   targ_state <= IDLE;

end tabort_transfer;





----------------------------------------------------------------------------------------------------

-- Procedure for Retry

-- write 5 data phases and then retry

----------------------------------------------------------------------------------------------------

procedure retry_transfer is

variable index : integer;

variable count : integer := 0;

begin



   targ_busy <= true;

   targ_state <= RETRY_XFER;

   index := vec2int(AD_in(27 downto 2));



   print("[HOST TARGET] Performing Retry...");

   print("    Address: " & vec2hstr(AD_in,8));



   -- enable driving of target signals

   targ_oe <= '1';



   -- wait for IRDY_N

   while To_X01(IRDY_N) = '1' loop

   	wait until rising_edge(CLK);

   end loop;



   -- assert DEVSEL_N

   reg_devsel_n <= '0';

   wait until rising_edge(CLK);



	-- assert TRYDY_N

   reg_trdy_n <= '0';

   targ_cbe <= CBE;



   -- if read cycle

   if read then

   	targ_ad_oe <= '1';

      wait for Tdel;

   	print("    Reading from host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   	targ_par_oe <= '1';

   else

      wait for Tdel;

   	print("    Writing to host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   end if;



	index := index + 1;

	-- loop 5 times or until frame asserted

	while count < 4 and To_X01(FRAME_N) = '0' and To_X01(IRDY_N) = '0' loop

      targ_cbe <= CBE;

      if read then

         wait for Tdel;

      else

         wait for Tdel;

      end if;

      print("    Data Transferred: " & vec2hstr(AD,8));

      wait until rising_edge(CLK);

      count := count + 1;

  		index := index + 1;

   end loop;



   -- if frame not asserted after 5 data phases

   if To_X01(FRAME_N) = '0' then

      -- assert retry signals

      reg_devsel_n <= '0';

      reg_trdy_n <= '1';

      reg_stop_n <= '0';

      targ_cbe <= CBE;

      targ_ad_oe <= '0';

   	wait until rising_edge(CLK);

      wait until rising_edge(CLK);

   end if;



   -- de-assert signals

   reg_devsel_n <= '1';

  	reg_trdy_n <= '1';

   reg_stop_n <= '1';

  	targ_ad_oe <= '0';

   targ_par_oe <= '0';

   wait until rising_edge(CLK);



   -- disable driving of target signals

   targ_oe <= '0';



   print("[HOST TARGET] End of Retry" & LF);

   targ_busy <= false;

   targ_state <= IDLE;

end retry_transfer;





----------------------------------------------------------------------------------------------------

-- Procedure for Partial transfer

-- Disconnect with data after One data phase

----------------------------------------------------------------------------------------------------

procedure partial_transfer is

variable index : integer;

begin



   targ_busy <= true;

   targ_state <= PARTIAL_XFER;

   index := vec2int(AD_in(27 downto 2));



   print("[HOST TARGET] Performing Partial Transfer..." & LF);

   print("    Start Address: " & vec2hstr(AD_in,8));



   -- enable driving of target signals

   targ_oe <= '1';



   -- wait for IRDY_N

   if To_X01(IRDY_N) = '1' then

      wait until To_X01(IRDY_N) = '0';

   end if;



   -- assert DEVSEL_N

   reg_devsel_n <= '0';

   wait until rising_edge(CLK);



	-- assert TRYDY_N

   reg_trdy_n <= '0';

   targ_cbe <= CBE;

   reg_stop_n <= '0';



   -- if read cycle

   if read then

   	targ_ad_oe <= '1';

      wait for Tdel;

   	print("    Reading from host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   	targ_par_oe <= '1';

   else

      wait for Tdel;

   	print("    Writing to host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   end if;



   -- de-assert TRDY_N

   reg_trdy_n <= '1';

   targ_ad_oe <= '0';

   targ_par_oe <= '1';

   wait until rising_edge(CLK);



   -- disable driving PAR

   reg_devsel_n <= '1';

   reg_stop_n <= '1';

   targ_par_oe <= '0';

   wait until rising_edge(CLK);



   -- disable driving of target signals

   targ_oe <= '0';



   print("[HOST TARGET] End of Partial Transfer" & LF);

   targ_busy <= false;

   targ_state <= IDLE;

end partial_transfer;





----------------------------------------------------------------------------------------------------

-- Procedure for full transfer

----------------------------------------------------------------------------------------------------

procedure full_transfer is

constant cMODULUS      : integer := 4;      	-- modulus amount

constant cRESULT       : integer := 3;      	-- test for mod function result

constant cINITIAL      : integer := 1;      	-- initial value of "count"



variable count    : integer;           	  	-- counter

variable data     : std_logic_vector(31 downto 0);

variable index : integer;

begin



   targ_busy <= true;

   index := vec2int(AD_in(27 downto 2));



	if gen_perr = '1' then

		print("[HOST TARGET] Performing Transfer with Parity Error...");

   	targ_state <= PERR_XFER;

   else

   	print("[HOST TARGET] Performing Full Transfer..." & LF);

   	targ_state <= FULL_XFER;

   end if;

   print("    Start Address: " & vec2hstr(AD_in,8));



   -- enable driving of target signals

   targ_oe <= '1';



   -- wait for IRDY_N

   while To_X01(IRDY_N) = '1' loop

   	wait until rising_edge(CLK);

   end loop;



   -- assert DEVSEL_N

   count := cINITIAL;

   reg_devsel_n <= '0';

   wait until rising_edge(CLK);



   -- assert TRYDY_N

   reg_trdy_n <= '0';

  	targ_cbe <= CBE;



   -- if read cycle

   if read then

   	targ_ad_oe <= '1';

      wait for Tdel;

   	print("    Reading from host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   	targ_par_oe <= '1';

   else

      wait for Tdel;

   	print("    Writing to host...");

   	print("    Data Transferred: " & vec2hstr(AD,8));

   	wait until rising_edge(CLK);

   end if;



	index := index + 1;

   while To_X01(FRAME_N) = '0' and To_X01(IRDY_N) = '0' loop

		if INSERT_WAITS then

         if count mod cMODULUS = cRESULT then

            reg_trdy_n <= '1';

         	wait until rising_edge(CLK);

            print("    Wait inserted");

         else

            reg_trdy_n <= '0';

            if read then

      			wait for Tdel;

            else

      			wait for Tdel;

            end if;

            print("    Data Transferred: " & vec2hstr(AD,8));

            wait until rising_edge(CLK);

         	index := index + 1;

         end if;

         count := count + 1;

      else

        	targ_cbe <= CBE;

         if read then

      		wait for Tdel;

         else

      		wait for Tdel;

         end if;

         print("    Data Transferred: " & vec2hstr(AD,8));

         wait until rising_edge(CLK);

         index := index + 1;

		end if;

   end loop;



   -- if trdy = 1, ie wait was inserted

   -- then finish last data phase

   if reg_trdy_n = '1' then

      reg_trdy_n <= '0';

      if read then

         wait for Tdel;

      else

         wait for Tdel;

      end if;

      print("    Data Transferred: " & vec2hstr(AD,8));

      wait until rising_edge(CLK);

   end if;



   reg_devsel_n <= '1';

   reg_trdy_n <= '1';

   targ_cbe <= CBE;

   targ_ad_oe <= '0';



   -- disable driving PAR

   wait until rising_edge(CLK);

   targ_par_oe <= '0';



   -- disable driving of target signals

   targ_oe <= '0';



   if gen_perr = '1' then

   	print("[HOST TARGET] End of Parity Error tranfer" & LF);

   else

   	print("[HOST TARGET] End of Full Transfer" & LF);

   end if;

   targ_busy <= false;

   targ_state <= IDLE;

end full_transfer;





----------------------------------------------------------------------------------------------------

begin  -- Main Target Loop

----------------------------------------------------------------------------------------------------

   wait on CLK, RESET_N;



   if END_SIM then

      targ_state <= FINISH;

      targ_busy  <= false;

      targ_oe  <= '0';

      wait;

   else

      if RESET_N = '0' then

         targ_oe      <= '0';

         targ_par_oe  <= '0';

         targ_ad_oe   <= '0';

         targ_ad      <= (others =>'0');

         targ_cbe_oe  <= '0';

         targ_cbe     <= (others =>'0');

         reg_trdy_n   <= '1';

         reg_stop_n   <= '1';

         reg_devsel_n <= '1';

         targ_state <= RESET;

         targ_busy  <= false;



         -- wait until reset de-asserts

         wait until RESET_N = '1';

         targ_state <= IDLE;



      elsif rising_edge(CLK) then

         if To_X01(FRAME_N) = '0' and To_X01(IRDY_N) = '1' then

            AD_in := AD;

            cbe_in := CBE;



				-- check if host is being addressed

            if AD_in >= HOST_BASE then

               -- determine if read or write cycle

               if cbe_in = IO_READ  or cbe_in = MEM_READ or cbe_in = CFG_READ or

                  cbe_in = READ_MULT or cbe_in = READ_LINE then

                  read := true;

               else

                  read := false;

               end if;



					case TARGET_CMD is

                  when TARGET_ABORT =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     tabort_transfer;

                  when MASTER_ABORT =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     mabort_transfer;

                  when PARITY_ERROR  =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     gen_perr <= '1';

                     full_transfer;

                     gen_perr <= '0';

                  when PARTIAL_BURST =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     -- do partial transfer only 50% of the time

                     if rnd_num < 0.5 then

                     	partial_transfer;

                     else

                        full_transfer;

                     end if;

                  when RETRY_BURST =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     retry_transfer;

                  when others =>

                     targ_ack <= true;

                     wait for 15 ps;

                     targ_ack <= false;

                     full_transfer;

               end case;

            end if;

         end if;

      end if;

   end if;

end process;





----------------------------------------------------------------------------------------------------

-- Random Number Generator Process

----------------------------------------------------------------------------------------------------

RANDOM_NUMBER : process

variable seed : real := 17.0;

variable data : real;

begin

   while not END_SIM loop

      RANDOM(seed, data);

      rnd_num <= data;

      wait for 10 ns;

   end loop;

end process;





----------------------------------------------------------------------------------------------------

end PCI_Host_arch;  -- end architecture

----------------------------------------------------------------------------------------------------

