//--------------------------------------------------------------------------------------------------

//             :

// File name   : SDRAM_PCI.V

//             :

// Title       : PCI portion of the S2PCI SDRAM controller testbench

//             :

// Comments    :

//             :

//             :

// Conventions : - Port names are 'UPPER' case.

//             : - Internal signals and variables are 'lower' case.

//             : - Active low signals are identified with '_n' or '_N'.

//             : - instances begin with lower case 'i', rest upper case

//             : - parameters begin with lower case 'p', rest upper case

//             : - tasks begin with lower case 't', rest upper case

//             :

//             : Tab setting = 3

//             : Right Margin = 100

//             :

// Limitations : For use in testbench only

//             :

// Developers  : Majid Kalsi

//             :

//--------------------------------------------------------------------------------------------------

// Revision History :

//--------------------------------------------------------------------------------------------------

//   Version No | Author    | Changes Made:                                            | Mod. Date:

//     v1.0     | da        | First Written          												| 27/08/01

//--------------------------------------------------------------------------------------------------





//------------------------------------------------------------------------------

//  Include Files

//------------------------------------------------------------------------------

`include "PCI_device_defines.v"



//------------------------------------------------------------------------------

//  Module declaration

//------------------------------------------------------------------------------

`timescale 1 ns/1 ps





module SDRAM_TB_PCI

   (

      // PCI signals

	AD,

   	CBE,

   	PAR,

   	FRAME_N,

   	TRDY_N,

   	IRDY_N,

   	STOP_N,

   	DEVSEL_N,

   	dut_IDSEL,

    dut_INTR_A,

   	PERR_N,

   	SERR_N,

   	dut_REQ_N,

   	dut_GNT_N,

   	RST_N,

   	PCLK,



      // Testbench Control Signals

      // command/control interface

   	ICMD,                   // Initiator Command

   	ICMD_ACK,               // Initiator Command Acknowledge

   	ISTATE,                 // Initiator State

   	IBUSY,                  // Initiator Busy

   	TCMD,                   // Target Command

   	TCMD_ACK,               // Target Command Acknowledge

   	TSTATE,                 // Target State

   	TBUSY,                  // Target Busy

   	END_SIM,                // End of Simulation Flag



   	// Initial Latency Control

   	IRDY_INITIAL_LATENCY,   // IRDY Initial latency, min 0, max 7

   	TRDY_INITIAL_LATENCY,   // TRDY Initial latency, min 0, max 15



   	// Wait State Control, see description above

   	INSERT_WAITS,           // Insert Wait States Flag

   	INITIAL_VALUE,          // initial value of "count"

   	TERM_CNT,               // termination count

   	MOD_AMT,                // modulus amount

   	RESULT,                 // test for mod function result

   	DURATION               // duration of waits, min 1, max 7

   );





//-----------------------------------------------

//  Port declarations

//-----------------------------------------------



// PCI



inout  			PAR;

inout  			FRAME_N;

inout  			TRDY_N;

inout  			IRDY_N;

inout  			STOP_N;

inout  			DEVSEL_N;

output  		dut_IDSEL;

input 			dut_INTR_A;

inout  			PERR_N;

inout  			SERR_N;

input 			dut_REQ_N;

output  		dut_GNT_N;

inout  			RST_N;

output  		PCLK;

inout  [31:0]	AD;

inout  [3:0]	CBE;



// Host Control



input [3:0] ICMD;

output ICMD_ACK;

output [99:0]ISTATE;

output IBUSY;

input [3:0] TCMD;

output TCMD_ACK;

output [99:0]TSTATE;

output TBUSY;

input END_SIM;



// Initial Latency Control

input [3:0]IRDY_INITIAL_LATENCY;

input [3:0]TRDY_INITIAL_LATENCY;



// Wait State Control, see description above

input INSERT_WAITS;

input [3:0] INITIAL_VALUE;

input [3:0] TERM_CNT;

input [3:0] MOD_AMT;

input [3:0] RESULT;

input [3:0] DURATION;





tri1 REQ64_N;

tri1 ACK64_N;

tri1 FRAME_N;

tri1 IRDY_N;

tri1 DEVSEL_N;

tri1 TRDY_N;

tri1 STOP_N;

tri1 [3:0] CBE;

tri1 [31:0] AD;

tri1 PAR;

tri1 PERR_N;

tri1 SERR_N;

tri1 PAR64;



//tri1



// PCI Monitor signals

wire [(30*`SIZEOFCHAR - 1):0] INIT_XFER;

wire [(30*`SIZEOFCHAR - 1):0] TARG_XFER;

wire [(30*`SIZEOFCHAR - 1):0] CYCLE;





wire BUSY;



parameter GND = 1'b0;

wire gnd_wire;

assign gnd_wire = GND;



reg  CLK;

wire  dh_REQ_N; 	// PCI request signal for host

wire  dh_GNT_N;   // PCI grant signal for host

wire  dh_IDSEL;   // PCI configuration select for host



// define IDSELs

assign dh_IDSEL =  ((FRAME_N == 0) && (CBE[3:1] == 3'b101) && AD[31])  ? 1'b1 : 1'b0;

assign dut_IDSEL = ((FRAME_N == 0) && (CBE[3:1] == 3'b101) && AD[29]) ? 1'b1 : 1'b0;



// define CLK

initial CLK = 0;

always #(15) CLK = !CLK;

// 20 = 25 MHz

// 15 = 33 MHz

// 7.5 = 66 MHZ

// 30 = 16.6 MHz



// define BUSY

assign BUSY = IBUSY;

assign PCLK = CLK;





// instantiate PCI ARBITER

DUMB_ARBITER #(2) iARBITER

(

   .CLK(CLK),

   .REQ_N({dh_REQ_N, dut_REQ_N}),

   .GNT_N({dh_GNT_N, dut_GNT_N}),

   .FRAME_N(FRAME_N),

   .IRDY_N(IRDY_N)

);





// instantiate HOST

PCI_DEVICE iHOST

(

   // command/control interface

   .ICMD(ICMD),

   .ICMD_ACK(ICMD_ACK),

   .ISTATE_out(ISTATE),

   .IBUSY(IBUSY),

   .TCMD(TCMD),

   .TCMD_ACK(TCMD_ACK),

   .TSTATE_out(TSTATE),

   .TBUSY(TBUSY),

   .END_SIM(END_SIM),



   // Initial Latency Control

   .IRDY_INITIAL_LATENCY(IRDY_INITIAL_LATENCY),

   .TRDY_INITIAL_LATENCY(TRDY_INITIAL_LATENCY),



   // Wait State Control

   .INSERT_WAITS(INSERT_WAITS),

   .INITIAL_VALUE(INITIAL_VALUE),

   .TERM_CNT(TERM_CNT),

   .MOD_AMT(MOD_AMT),

   .RESULT(RESULT),

   .DURATION(DURATION),



   // PCI signals

   .CLK(CLK),

   .RST_N(RST_N),

   .INTR_A(dut_INTR_A),

   .REQ_N(dh_REQ_N),

   .GNT_N(dh_GNT_N),

   .REQ64_N (),

   .ACK64_N (gnd_wire),

   .FRAME_N(FRAME_N),

   .IRDY_N(IRDY_N),

   .IDSEL(dh_IDSEL),

   .DEVSEL_N(DEVSEL_N),

   .TRDY_N(TRDY_N),

   .STOP_N(STOP_N),

   .CBE(CBE),

   .AD(AD),

   .PAR(PAR),

   .PERR_N(PERR_N),

   .SERR_N(SERR_N),

   .PAR64 (gnd_wire)



);



// instantiate PCI Bus Monitor

PCI_MONITOR iPCI_MONITOR

(

   .CLK(CLK),

   .FRAME_N(FRAME_N),

   .IRDY_N(IRDY_N),

   .DEVSEL_N(DEVSEL_N),

   .TRDY_N(TRDY_N),

   .STOP_N(STOP_N),

   .CBE(CBE[3:0]),

   .INIT_XFER(INIT_XFER),

   .TARG_XFER(TARG_XFER),

   .CYCLE(CYCLE)

);





endmodule

